@@ -8,6 +8,7 @@ import shutil
 import subprocess
 import sys
 import tempfile
+import uuid
 from xml.dom import minidom
 
 from metakernel import MetaKernel, ProcessMetaKernel, REPLWrapper, u
@@ -18,7 +19,7 @@ from . import __version__
 
 
 STDIN_PROMPT = '__stdin_prompt>'
-STDIN_PROMPT_REGEX = re.compile(r'\A.+?%s' % STDIN_PROMPT)
+STDIN_PROMPT_REGEX = re.compile(r'\A.+?%s|debug> ' % STDIN_PROMPT)
 
 
 class OctaveKernel(ProcessMetaKernel):
@@ -51,7 +52,8 @@ class OctaveKernel(ProcessMetaKernel):
         self._octave_engine = OctaveEngine(plot_settings=self.plot_settings,
                                            error_handler=self.Error,
                                            stdin_handler=self.raw_input,
-                                           stream_handler=self.Print)
+                                           stream_handler=self.Print,
+                                           logger=self.log)
         return self._octave_engine
 
     def makeWrapper(self):
@@ -84,6 +86,9 @@ class OctaveKernel(ProcessMetaKernel):
         # Ignore standalone input hook displays.
         if (args[0].strip() == STDIN_PROMPT):
             return
+        if (args[0].startswith(STDIN_PROMPT)):
+            args = list(args)
+            args[0] = args[0].replace(STDIN_PROMPT, '')
         super(OctaveKernel, self).Print(*args, **kwargs)
 
     def raw_input(self, text):
@@ -107,7 +112,9 @@ class OctaveKernel(ProcessMetaKernel):
 class OctaveEngine(object):
 
     def __init__(self, error_handler=None, stream_handler=None,
-                 stdin_handler=None, plot_settings=None):
+                 stdin_handler=None, plot_settings=None,
+                 logger=None):
+        self.logger = logger
         self.executable = self._get_executable()
         self.repl = self._create_repl()
         self.error_handler = error_handler
@@ -155,11 +162,20 @@ class OctaveEngine(object):
         """Evaluate code using the engine.
         """
         stream_handler = None if silent else self.stream_handler
+        if self.logger:
+            self.logger.debug('Octave eval:')
+            self.logger.debug(code)
         try:
-            return self.repl.run_command(code.rstrip(),
+            resp = self.repl.run_command(code.rstrip(),
                                          timeout=timeout,
                                          stream_handler=stream_handler,
                                          stdin_handler=self.stdin_handler)
+            resp = resp.replace(STDIN_PROMPT, '')
+            if self.logger and resp:
+                self.logger.debug(resp)
+            return resp
+        except KeyboardInterrupt:
+            return self._interrupt(True)
         except Exception as e:
             if self.error_handler:
                 self.error_handler(e)
@@ -194,7 +210,6 @@ class OctaveEngine(object):
         make_figs = '_make_figures("%s", "%s", "%s", %d, %d, %d)'
         make_figs = make_figs % (plot_dir, fmt, name, wid, hgt, res)
         resp = self.eval(make_figs, silent=True)
-        print(resp)
         if resp and 'error:' in resp:
             if self.error_handler:
                 self.error_handler(resp)
@@ -229,10 +244,9 @@ class OctaveEngine(object):
 
     def _startup(self, plot_settings):
         cwd = os.getcwd().replace(os.path.sep, '/')
-        self.eval('more off; source ~/.octaverc; cd("%s");' % cwd)
+        self.eval('more off; source ~/.octaverc; cd("%s");' % cwd, silent=True)
         here = os.path.realpath(os.path.dirname(__file__))
         self.eval('addpath("%s")' % here.replace(os.path.sep, '/'))
-        self.eval('add_input_event_hook(@_input_hook);')
         available = self.eval('available_graphics_toolkits', silent=True)
         if 'gnuplot' not in available:
             msg = ('May not be able to display plots properly '
@@ -302,9 +316,54 @@ class OctaveEngine(object):
 
         repl = REPLWrapper(cmd, orig_prompt, change_prompt,
                            stdin_prompt_regex=STDIN_PROMPT_REGEX)
-        repl.linesep = '\n'
+        if os.name == 'nt':
+            repl.child.crlf = '\n'
+        repl.interrupt = self._interrupt
         return repl
 
+    def _interrupt(self, silent=False):
+        if (os.name == 'nt'):
+            if self.logger:
+                msg = 'Cannot interrupt Octave kernel on Windows'
+                self.logger.warning(msg)
+            return self._interrupt_expect(silent)
+        return REPLWrapper.interrupt(self.repl)
+
+    def _interrupt_expect(self, silent):
+        repl = self.repl
+        child = repl.child
+        expects = [repl.prompt_regex, child.linesep]
+        expected = uuid.uuid4().hex
+        repl.sendline('disp("%s");' % expected)
+        if repl.prompt_emit_cmd:
+            repl.sendline(repl.prompt_emit_cmd)
+        lines = []
+        while True:
+            # Prevent a keyboard interrupt from breaking this up.
+            while True:
+                try:
+                    pos = child.expect(expects)
+                    break
+                except KeyboardInterrupt:
+                    pass
+            if pos == 1:  # End of line received
+                line = child.before
+                if silent:
+                    lines.append(line)
+                else:
+                    self.stream_handler(line)
+            else:
+                line = child.before
+                if line.strip() == expected:
+                    break
+                if len(line) != 0:
+                    # prompt received, but partial line precedes it
+                    if silent:
+                        lines.append(line)
+                    else:
+                        self.stream_handler(line)
+        return '\n'.join(lines)
+
     def _get_executable(self):
         """Find the best octave executable.
         """
